"""
qc_plot.py
Functions to plot preliminary QC plots. Pre-processing using read_data functions is required
"""
import matplotlib.pyplot as plt
import spatialdata_plot # used for rendering images method
import numpy as np
import scanpy as sc

def compute_qc_metrics(sdata, bin_size = 2, mito_prefix="MT-"):
    """
    Compute standard QC metrics using Scanpy and add log-transformed total counts.
    Coupled with plot_qc_metrics below
    Reference: https://scanpy.readthedocs.io/en/stable/generated/scanpy.pp.calculate_qc_metrics.html

    Adds the following columns to sdata.tables["square_002um"].obs:
    - 'total_counts': sum of raw counts per spot
    - 'n_genes_by_counts': number of genes per spot
    - 'pct_counts_mt': percent of counts from mito genes
    - 'log_total_counts': log1p of total counts

    Args:
    sdata: the SpatialData 
    bin_size: bin resolution in um (must be from 2,8,16). default to 2um
    mito_prefix: prefix used to identify mitochondrial genes (default is "MT-")

    Return:
    Modified SpatialData object with updated QC metrics.
    """
    if bin_size not in [2, 8, 16]:
        raise ValueError("bin_size must be one of: 2, 8, 16") # double check bin size
    bin_str = f"{bin_size:03d}um" # padding
    table_key = f"square_{bin_str}"
    adata = sdata.tables[table_key]
    # Add boolean column (to identify MT genes) to var
    adata.var["mt"] = adata.var_names.str.startswith(mito_prefix)

    # Calculate QC metrics using scanpy
    sc.pp.calculate_qc_metrics(
        adata,
        qc_vars={"mt": adata.var["mt"]},
        percent_top=None,
        log1p=False,
        inplace=True
    )

    # Add log-transformed total counts
    adata.obs["log_total_counts"] = np.log1p(adata.obs["total_counts"].values)

    return sdata


def plot_qc_metrics(sdata, id, metric, bin_size=2, cmap="viridis", vmin=None, vmax=None, save_path=None):
    """
    Plot a single QC metric (e.g., total counts, % mito) at 2um resolution.

    Args:
    sdata: the SpatialData object
    id: ID of the VisiumHD dataset (ex. "H1-TRH234F_D1")
    metric: name of the QC metric in sdata.tables["square_002um"].obs to visualize.
            Examples: "total_counts", "log_total_counts", "n_genes_by_counts", "pct_counts_mt" from compute function
    bin_size: bin resolution in um (must be from 2,8,16). default to 2um
    cmap: matplotlib colormap name (default: "viridis")
    vmin: (optional) lower limit for color scale
    vmax: (optional) upper limit for color scale
    save_path: (optional) the path to save figure. If None (default), shows plot interactively.
    """
    from spatialdata_plot.pl.utils import set_zero_in_cmap_to_transparent

    if bin_size not in [2, 8, 16]:
        raise ValueError("bin_size must be one of: 2, 8, 16")

    bin_str = f"{bin_size:03d}um"
    table_key = f"square_{bin_str}"
    bin_key = f"{id}_{table_key}"
    adata = sdata.tables[table_key]

    # Validate metric exists
    if metric not in adata.obs.columns:
        raise ValueError(f"Metric '{metric}' not found in sdata.tables['{table_key}'].obs")

    # Make zero-transparent colormap for sparse data
    custom_cmap = set_zero_in_cmap_to_transparent(cmap=cmap)

    # plot
    sdata.pl.render_shapes(
        bin_key,
        color=metric,
        cmap=custom_cmap,
        vmin=vmin,
        vmax=vmax,
        method = "datashader",
        datashader_reduction = "max" # to preserve max of the data for colorscale
    ).pl.show(
        coordinate_systems="global",
        title=f"{bin_size}um QC: {metric}",
        figsize=(10, 10),
        save=save_path
    )

def plot_gene_exp(sdata, id, gene_name):
    """
    Plots gene expression of a particular gene across 16um and 8um bins
    If no expression in a particular bin, the bin is transparent
    Otherwise, follows the scale in the generated figure

    Args:
    sdata: the SpatialData object
    id: the ID of the VisiumHD dataset, ex: 'H1-TRH234F_D1'
    gene_name: the gene of interest
    """
    # make the color scale
    # sparsity - visualize non-zero entries only, use FULL-RES image as background
    from spatialdata_plot.pl.utils import set_zero_in_cmap_to_transparent
    # display the areas where no expression is detected as transparent
    new_cmap = set_zero_in_cmap_to_transparent(cmap="viridis")

    for bin_size in [16, 8]:
        sdata.pl.render_shapes(
            f"{id}_square_{bin_size:03}um",
            color=gene_name,
            cmap=new_cmap
        ).pl.show(
            coordinate_systems="global",
            title=f"bin_size={bin_size}µm",
            figsize=(10, 10)
        )


def plot_full_image(sdata, image_key):
    """
    Plots the full image for reference.

    Args:
    sdata: the SpatialData object
    image_key: the key to the image plotted. Ex: "H1-TRH234F_D1_full_image"
    """
    fig, ax = plt.subplots(figsize=(5, 5))  # Single plot   
    # Plot the full resolution image
    sdata.pl.render_images(image_key).pl.show(ax=ax, title="Full Image")
    plt.tight_layout()
    plt.show()


def plot_cropped_image(sdata, xmin, ymin, xmax, ymax, image_key):
    """
    Plot a cropped region of the image.

    Args:
    sdata: the SpatialData object
    xmin: the min x coordinate
    ymin: the min y coordinate
    xmax: the max x coordinate
    ymax: the max y coordinate
    image_key: the key to the image plotted. Ex: "H1-TRH234F_D1_full_image"
    """
    from spatialdata import bounding_box_query
    fig, ax = plt.subplots(figsize=(5, 5))  # Single plot

    # Define crop function
    crop0 = lambda x: bounding_box_query(
        x, 
        min_coordinate=[xmin, ymin], 
        max_coordinate=[xmax, ymax], 
        axes=("x", "y"), 
        target_coordinate_system="global"
    )

    sdata_cropped = crop0(sdata)

    # Apply crop and plot only the full image
    sdata_cropped.pl.render_images(image_key).pl.show( # type: ignore
        ax=ax, 
        title="Full Image (Cropped)", 
        coordinate_systems="global"
    )
    plt.tight_layout()
    plt.show()


"""
read_data.py
This file defines functions to read in VisiumHD data and convert to appropriate forms for analyses.
"""
from spatialdata_io import visium_hd
import spatialdata as sd
import shutil
import os
import copy

def to_spatialdata(raw_folder_path, id, fullres_image_path = 'outs/spatial/tissue_hires_image.png'):
    """
    This function reads in a raw Visium HD dataset, write to Zarr format
    then read in the Zarr data, and convert to a SpatialData object.
    (the Zarr directory is temporarily generated and then removed before returning)
    CAREFUL: WRITE access is needed to the original VisiumHD folder because of spatialdata limitation

    Args: 
    raw_folder_path: the path to the raw Visium HD data folder (outermost)
    id: the ID of the VisiumHD dataset, ex: 'H1-TRH234F_D1'
    fullres_image_path: the relative path to the tissue_hires_image in VisiumHD output. default to the one in outs folder

    Return:
    sdata, the resulting SpatialData object for downstream processing
    """
    # create a temporary address for the zarr file
    zarr_path = "temp.zarr" 
    # spatialdata library limitation: feature_slice.h5 needs to appear in the outermost directory
    # with its name changed to id_feature_slice.h5 (ex: H1-TRH234F_D1_feature_slice.h5)
    # copy feature_slice.h5 in the outs/ folder to the raw_folder_path, and rename to id_feature_slice.h5 where id the passed in param
    dst_feature_slice = os.path.join(raw_folder_path, f"{id}_feature_slice.h5")
    src_feature_slice = os.path.join(raw_folder_path, "outs", "feature_slice.h5")
    shutil.copyfile(src_feature_slice, dst_feature_slice)

    # read HD data, write/read from Zarr
    sdata = visium_hd(raw_folder_path, dataset_id = id, fullres_image_file = fullres_image_path)
    sdata.write(zarr_path)
    sdata = sd.read_zarr(zarr_path)
    sdata = copy.deepcopy(sdata) # deep copy to detach from Zarr
    sdata.path = None # reset Zarr path to avoid issues
    # summary
    print(sdata)
    # make the var names unique
    for table in sdata.tables.values():
        table.var_names_make_unique()
    # cleanup
    shutil.rmtree(zarr_path)
    os.remove(dst_feature_slice)
    return sdata


"""
write_data.py
This file defines functions to write updated spatialdata objects back to the original raw VisiumHD output form.
Some libraries, like FICTURE, want the original output folder as the input.
"""
import os
import pandas as pd
from scipy.io import mmwrite
import gzip

def write_2um_filtered_counts(sdata, raw_folder_path, new_name = "matrix_destriped.mtx.gz", overwrite = False):
    """
    Function that writes the filtered/destriped 2um counts back to the original HD folder, with a new file name.
    This is primarily intended to run after the destriping command ONLY.
    NOTE: this assumes gene list and barcodes unmodified.

    Args:
    sdata: the updated spatialdata object
    raw_folder_path: the raw VisiumHD folder (outermost directory)
    new_name: the file name for the new matrix
    overwrite: if False, raise error if overwriting happens
    """
    adata = sdata.tables["square_002um"] # extract data
    # define folders and file paths
    matrix_dir = os.path.join(
        raw_folder_path,
        "outs/binned_outputs/square_002um/filtered_feature_bc_matrix"
    )
    barcodes_path = os.path.join(matrix_dir, "barcodes.tsv.gz")
    features_path = os.path.join(matrix_dir, "features.tsv.gz")
    new_matrix_path = os.path.join(matrix_dir, new_name)

    # Load and validate barcode/gene information
    old_barcodes = pd.read_csv(barcodes_path, sep="\t", header=None).iloc[:, 0]
    old_genes = pd.read_csv(features_path, sep="\t", header=None).iloc[:, 1]

    assert (adata.obs_names == old_barcodes.values).all(), "Barcodes do not match!"
    assert (adata.var_names == old_genes.values).all(), "Gene names do not match!"

    # overwrite check
    if os.path.exists(new_matrix_path) and not overwrite:
        raise FileExistsError(
            f"{new_matrix_path} already exists."
        )

    # Write matrix to new file
    # Support both gz-compressed and plain matrix files
    if new_matrix_path.endswith(".gz"):
        with gzip.open(new_matrix_path, "wb") as new_file:
            mmwrite(new_file, adata.X)
    else:
        mmwrite(new_matrix_path, adata.X)


"""
destripe.py
Includes function to reduce the striping effect of VisiumHD data.
Pre-processing using read_data functions is required.
Some functions are incorporated from bin2cell package.
"""
import numpy as np
import scanpy as sc

def destripe_b2c(sdata, quantile = 0.99, only_2um=True, plot=False):
    """
    function to destripe VisiumHD data using the method provided by bin2cell package.
    Note: bin2cell works on AnnData objects, which need to be extracted from spatialdata object.

    Correct the raw counts of the input object for known variable width of 
    VisiumHD 2um bins. Scales the total UMIs per bin on a per-row and 
    per-column basis, dividing by the specified ``quantile``. The resulting 
    value is stored in ``.obs[factor_key]``, and is multiplied by the 
    corresponding total UMI ``quantile`` to get ``.obs[adjusted_counts_key]``.

    Args:
    sdata: the spatialdata object
    quantile: the quantile needed for destriping (default at 99% by b2c)
    only2um: a boolean parameter to indicate whether only 2um bins are de-striped. default to True.
    (Future work to add de-striping to 8um and 16um bins)
    plot: whether to generate summary plots (not yet implemented)

    Return:
    sdata: the resulting spatialdata object
    """
    def destripe(adata, quantile=0.99, counts_key="n_counts", factor_key="destripe_factor",
                 adjusted_counts_key="n_counts_adjusted", adjust_counts=True):
        """
        Perform destriping calculations.
        Args:
        adata: 2um bin output (AnnData)
        quantile: quantile required for destriping (default = 0.99)
        counts_key: name of column with raw counts per bin (default = n_counts)
        factor_key: name of column to hold computed factor prior to reversing to count space (default = destripe_factor)
        adjusted_counts_key: name of column for storing destriped counts per bin (default = n_counts_adjusted)
        adjust_counts: boolean, whether to use computed adjusted count total to adjust counts in adata.X (default = True)
        """
        # validity check
        if counts_key not in adata.obs:
            raise KeyError(f"'{counts_key}' not found in adata.obs. Available columns: {adata.obs.columns.tolist()}")
        
        # Row-wise normalization
        row_q = adata.obs.groupby("array_row")[counts_key].quantile(quantile) # get quantile per row
        adata.obs[factor_key] = adata.obs[counts_key] / adata.obs["array_row"].map(row_q) # divide by quantile
        # clean-up bad values
        adata.obs[factor_key].replace([np.inf, -np.inf], np.nan, inplace=True)
        adata.obs[factor_key].fillna(1.0, inplace=True)

        # Column-wise normalization
        col_q = adata.obs.groupby("array_col")[factor_key].quantile(quantile)
        adata.obs[factor_key] /= adata.obs["array_col"].map(col_q)
        # clean-up bad values
        adata.obs[factor_key].replace([np.inf, -np.inf], np.nan, inplace=True)
        adata.obs[factor_key].fillna(1.0, inplace=True)
        
        # Global adjustment (global quantile * destripe factor)
        global_q = np.quantile(adata.obs[counts_key], quantile)
        adata.obs[adjusted_counts_key] = adata.obs[factor_key] * global_q
        # Adjust count matrix
        if adjust_counts:
            destripe_counts(adata, counts_key, adjusted_counts_key)
    
    def destripe_counts(adata, counts_key="n_counts", adjusted_counts_key="n_counts_adjusted"):
        """
        Scale each row (bin) of adata.X to have adjusted total counts.
        Args:
        adata: 2um bin output (AnnData)
        counts_key: name of column with raw counts per bin
        adjusted_counts_key: name of column storing the destriped counts per bin
        """
        #scanpy's utility function to make sure the anndata is not a view
        #if it is a view then weird stuff happens when you try to write to its .X
        sc._utils.view_to_actual(adata)
        # adjust count matrix
        scaling_factors = (adata.obs[adjusted_counts_key] / adata.obs[counts_key]).replace([np.inf, -np.inf], np.nan).fillna(1.0) # handle bad values (NaN, inf, -inf), define scaling factor = 1
        adata.X = adata.X.multiply(scaling_factors.values[:, None]) # matrix multiplication, edited for efficiency
        adata.obs[counts_key] = np.array(adata.X.sum(axis=1)).flatten() # edit n_counts based on destriped counts

    if only_2um: # only handling 2um
        key = "square_002um"
        if key not in sdata.tables: # double check if key is present
            raise KeyError(f"{key} not found in sdata.tables")
        adata = sdata.tables[key]
        adata.obs["n_counts"] = np.array(adata.X.sum(axis=1)).flatten() # compute n_counts
        destripe(adata, quantile=quantile)
        sdata.tables[key] = adata  # update modified table back

    return sdata